import os
from datetime import datetime
from itertools import product

import astropy
import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import GCRS, SkyCoord, get_sun
from astropy.time import Time
from astropy.wcs import WCS
from sunpy.coordinates import frames, get_earth

from punchbowl.data.wcs import (
    calculate_celestial_wcs_from_helio,
    calculate_helio_wcs_from_celestial,
    calculate_pc_matrix,
    extract_crota_from_wcs,
    load_trefoil_wcs,
)

_ROOT = os.path.abspath(os.path.dirname(__file__))


def test_sun_location():
    time_current = Time(datetime.utcnow())

    skycoord_sun = astropy.coordinates.get_sun(time_current)

    skycoord_origin = SkyCoord(0*u.deg, 0*u.deg,
                              frame=frames.Helioprojective,
                              obstime=time_current,
                              observer='earth')

    # with frames.Helioprojective.assume_spherical_screen(skycoord_origin.observer):
    skycoord_origin_celestial = skycoord_origin.transform_to(GCRS)

    # with frames.Helioprojective.assume_spherical_screen(skycoord_origin.observer):
    assert skycoord_origin_celestial.separation(skycoord_sun) < 1 * u.arcsec
    assert skycoord_origin.separation(skycoord_sun) < 1 * u.arcsec


def test_sun_location_gcrs():
    time_current = Time(datetime.utcnow())

    skycoord_sun = astropy.coordinates.get_sun(time_current)

    skycoord_origin = SkyCoord(0*u.deg, 0*u.deg,
                              frame=frames.Helioprojective,
                              obstime=time_current,
                              observer='earth')

    # with frames.Helioprojective.assume_spherical_screen(skycoord_origin.observer):
    skycoord_origin_helio = skycoord_sun.transform_to(frames.Helioprojective)

    # with frames.Helioprojective.assume_spherical_screen(skycoord_origin.observer):
    assert skycoord_origin_helio.separation(skycoord_sun) < 1 * u.arcsec
    assert skycoord_origin.separation(skycoord_sun) < 1 * u.arcsec

def test_extract_crota():
    pc = calculate_pc_matrix(-10*u.deg, [-1, 2])

    wcs_celestial = WCS(naxis=2)
    wcs_celestial.wcs.ctype = ("RA---AZP", "DEC--AZP")
    wcs_celestial.wcs.cunit = ("deg", "deg")
    wcs_celestial.wcs.cdelt = (-1, 2)
    wcs_celestial.wcs.crpix = (2048.5, 2048.5)
    wcs_celestial.wcs.crval = (0, 0)
    wcs_celestial.wcs.pc = pc
    assert np.allclose(extract_crota_from_wcs(wcs_celestial), -10 * u.deg)

def test_extract_crota_helio():
    pc = calculate_pc_matrix(10*u.deg, [0.0225, 0.0225])
    wcs_helio = WCS(naxis=2)
    wcs_helio.wcs.ctype = ("HPLN-AZP", "HPLT-AZP")
    wcs_helio.wcs.cunit = ("deg", "deg")
    wcs_helio.wcs.cdelt = (0.0225, 0.0225)
    wcs_helio.wcs.crpix = (2048.5, 2048.5)
    wcs_helio.wcs.crval = (0, 0)
    wcs_helio.wcs.pc = pc
    assert np.allclose(extract_crota_from_wcs(wcs_helio), 10 * u.deg)

def test_helio_to_celestial_wcs_many_points():
    for date_obs, crval1, crval2, crot in product(
            ["2021-03-20T00:00:00", "2021-01-20T00:00:00"],
            [-20, 10, 30],
            [-10, 0, 10],
            [-30, 0, 30] * u.deg):
        earth = get_earth(date_obs)
        wcs_helio = WCS({"CRVAL1": crval1,
                         "CRVAL2": crval2,
                         "CRPIX1": 2047.5,
                         "CRPIX2": 2047.5,
                         "NAXIS1": 4096,
                         "NAXIS2": 4096,
                         "CDELT1": 0.0225,
                         "CDELT2": 0.0225,
                         "CUNIT1": "deg",
                         "CUNIT2": "deg",
                         "CTYPE1": "HPLN-AZP",
                         "CTYPE2": "HPLT-AZP",
                         "PC1_1": np.cos(crot).value,
                         "PC1_2": -np.sin(crot).value,
                         "PC2_1": np.sin(crot).value,
                         "PC2_2": np.cos(crot).value,
                         "DATE-OBS": date_obs,
                         "MJD-OBS": Time(date_obs).mjd,
                         "DSUN_OBS": earth.radius.to_value(u.m),
                         "HGLN_OBS": earth.lon.to_value(u.deg),
                         "HGLT_OBS": earth.lat.to_value(u.deg),
                         "PV2_1": 0,
                         })

        wcs_celestial = calculate_celestial_wcs_from_helio(wcs_helio)

        npoints = 20
        xs, ys = np.meshgrid(
            np.linspace(0, wcs_helio.pixel_shape[0], npoints),
            np.linspace(0, wcs_helio.pixel_shape[1], npoints),
        )

        hp_coords = wcs_helio.pixel_to_world(xs, ys)
        # These come out claiming to be ICRS, but they're really GCRS (since it doesn't seem possible to tell a WCS
        # that it's GCRS)
        eq_coords = wcs_celestial.pixel_to_world(xs, ys)
        hp_to_eq_coords = hp_coords.transform_to('gcrs')

        eq_ra = eq_coords.ra.to_value(u.deg)
        hp_to_eq_ra = hp_to_eq_coords.ra.to_value(u.deg)
        straddles_wrap = np.nonzero(np.abs(eq_ra - hp_to_eq_ra) > 350)
        for r, c in zip(*straddles_wrap):
            if eq_ra[r, c] > 350:
                eq_ra[r, c] -= 360
            else:
                hp_to_eq_ra[r, c] -= 360

        np.testing.assert_allclose(eq_ra, hp_to_eq_ra, atol=1e-12)
        np.testing.assert_allclose(eq_coords.dec.to_value(u.deg), hp_to_eq_coords.dec.to_value(u.deg), atol=1e-12)

def test_celestial_to_helio_wcs_many_points():
    for date_obs, crval1, crval2, crot in product(
            ["2021-03-20T00:00:00", "2021-01-20T00:00:00"],
            [10, 90, 220],
            [-30, 0, 30],
            [-30, 0, 30] * u.deg):
        wcs_celestial = WCS({"CRVAL1": crval1,
                         "CRVAL2": crval2,
                         "CRPIX1": 2047.5,
                         "CRPIX2": 2047.5,
                         "NAXIS1": 4096,
                         "NAXIS2": 4096,
                         "CDELT1": -0.0225,
                         "CDELT2": 0.0225,
                         "CUNIT1": "deg",
                         "CUNIT2": "deg",
                         "CTYPE1": "RA---AZP",
                         "CTYPE2": "DEC--AZP",
                         "PC1_1": np.cos(crot).value,
                         "PC1_2": np.sin(crot).value,
                         "PC2_1": -np.sin(crot).value,
                         "PC2_2": np.cos(crot).value,
                         "DATE-OBS": date_obs,
                         "MJD-OBS": Time(date_obs).mjd,
                         "PV2_1": 0,
                         })

        wcs_helio, _ = calculate_helio_wcs_from_celestial(wcs_celestial)

        npoints = 20
        xs, ys = np.meshgrid(
            np.linspace(0, wcs_celestial.pixel_shape[0], npoints),
            np.linspace(0, wcs_celestial.pixel_shape[1], npoints),
        )

        hp_coords = wcs_helio.pixel_to_world(xs, ys)
        # These come out claiming to be ICRS, but they're really GCRS (since it doesn't seem possible to tell a WCS
        # that it's GCRS)
        eq_coords = wcs_celestial.pixel_to_world(xs, ys)
        # Make them bona-fide GCRS
        eq_coords = SkyCoord(eq_coords.ra, eq_coords.dec, frame='gcrs', obstime=date_obs)
        eq_to_hp_coords = eq_coords.transform_to('helioprojective')

        hp_tx = hp_coords.Tx.to_value(u.deg)
        eq_to_hp_tx = eq_to_hp_coords.Tx.to_value(u.deg)
        straddles_wrap = np.nonzero(np.abs(hp_tx - eq_to_hp_tx) > 350)
        for r, c in zip(*straddles_wrap):
            if hp_tx[r, c] > 170:
                hp_tx[r, c] -= 360
            else:
                eq_to_hp_tx[r, c] -= 360

        np.testing.assert_allclose(hp_tx, eq_to_hp_tx, atol=1e-12)
        np.testing.assert_allclose(hp_coords.Ty.to_value(u.deg), eq_to_hp_coords.Ty.to_value(u.deg), atol=1e-12)


def test_helio_to_celestial_wcs_many_points_3d():
    # The third dimension here is Stokes
    for date_obs, crval1, crval2, crot in product(
            ["2021-03-20T00:00:00", "2021-01-20T00:00:00"],
            [-20, 10, 30],
            [-10, 0, 10],
            [-30, 0, 30] * u.deg):
        earth = get_earth(date_obs)
        wcs_helio = WCS({"CRVAL1": crval1,
                         "CRVAL2": crval2,
                         "CRVAL3": 0,
                         "CRPIX1": 2047.5,
                         "CRPIX2": 2047.5,
                         "CRPIX3": 0,
                         "NAXIS1": 4096,
                         "NAXIS2": 4096,
                         "NAXIS3": 3,
                         "CDELT1": 0.0225,
                         "CDELT2": 0.0225,
                         "CDELT3": 1.0,
                         "CUNIT1": "deg",
                         "CUNIT2": "deg",
                         "CUNIT3": "",
                         "CTYPE1": "HPLN-AZP",
                         "CTYPE2": "HPLT-AZP",
                         "CTYPE3": "STOKES",
                         "PC1_1": np.cos(crot).value,
                         "PC1_2": -np.sin(crot).value,
                         "PC2_1": np.sin(crot).value,
                         "PC2_2": np.cos(crot).value,
                         "DATE-OBS": date_obs,
                         "MJD-OBS": Time(date_obs).mjd,
                         "DSUN_OBS": earth.radius.to_value(u.m),
                         "HGLN_OBS": earth.lon.to_value(u.deg),
                         "HGLT_OBS": earth.lat.to_value(u.deg),
                         "PV2_1": 0,
                         })

        wcs_celestial = calculate_celestial_wcs_from_helio(wcs_helio)

        npoints = 20
        xs, ys = np.meshgrid(
            np.linspace(0, wcs_helio.pixel_shape[0], npoints),
            np.linspace(0, wcs_helio.pixel_shape[1], npoints))
        zs = np.ones_like(xs)

        hp_coords = wcs_helio.pixel_to_world(xs, ys, zs)
        # These come out claiming to be ICRS, but they're really GCRS (since it doesn't seem possible to tell a WCS
        # that it's GCRS)
        eq_coords = wcs_celestial.pixel_to_world(xs, ys, zs)
        hp_to_eq_coords = hp_coords[0].transform_to('gcrs')

        eq_ra = eq_coords[0].ra.to_value(u.deg)
        hp_to_eq_ra = hp_to_eq_coords.ra.to_value(u.deg)
        straddles_wrap = np.nonzero(np.abs(eq_ra - hp_to_eq_ra) > 350)
        for r, c in zip(*straddles_wrap):
            if eq_ra[r, c] > 350:
                eq_ra[r, c] -= 360
            else:
                hp_to_eq_ra[r, c] -= 360

        np.testing.assert_allclose(eq_ra, hp_to_eq_ra, atol=1e-12)
        np.testing.assert_allclose(eq_coords[0].dec.to_value(u.deg), hp_to_eq_coords.dec.to_value(u.deg), atol=1e-12)
        np.testing.assert_array_equal(hp_coords[1], eq_coords[1])


def test_celestial_to_helio_wcs_many_points_3d():
    # The third dimension here is Stokes
    for date_obs, crval1, crval2, crot in product(
            ["2021-03-20T00:00:00", "2021-01-20T00:00:00"],
            [10, 90, 220],
            [-30, 0, 30],
            [-30, 0, 30] * u.deg):
        wcs_celestial = WCS({"CRVAL1": crval1,
                         "CRVAL2": crval2,
                         "CRVAL3": 0,
                         "CRPIX1": 2047.5,
                         "CRPIX2": 2047.5,
                         "CRPIX3": 0,
                         "NAXIS1": 4096,
                         "NAXIS2": 4096,
                         "NAXIS3": 3,
                         "CDELT1": -0.0225,
                         "CDELT2": 0.0225,
                         "CDELT3": 1.0,
                         "CUNIT1": "deg",
                         "CUNIT2": "deg",
                         "CUNIT3": "",
                         "CTYPE1": "RA---AZP",
                         "CTYPE2": "DEC--AZP",
                         "CTYPE3": "STOKES",
                         "PC1_1": np.cos(crot).value,
                         "PC1_2": np.sin(crot).value,
                         "PC2_1": -np.sin(crot).value,
                         "PC2_2": np.cos(crot).value,
                         "DATE-OBS": date_obs,
                         "MJD-OBS": Time(date_obs).mjd,
                         "PV2_1": 0,
                         })

        wcs_helio, _ = calculate_helio_wcs_from_celestial(wcs_celestial)

        npoints = 20
        xs, ys = np.meshgrid(
            np.linspace(0, wcs_celestial.pixel_shape[0], npoints),
            np.linspace(0, wcs_celestial.pixel_shape[1], npoints))
        zs = np.ones_like(xs)

        hp_coords = wcs_helio.pixel_to_world(xs, ys, zs)
        # These come out claiming to be ICRS, but they're really GCRS (since it doesn't seem possible to tell a WCS
        # that it's GCRS)
        eq_coords = wcs_celestial.pixel_to_world(xs, ys, zs)
        eq_coords = SkyCoord(eq_coords[0].ra, eq_coords[0].dec, frame='gcrs', obstime=date_obs), eq_coords[1]
        eq_to_hp_coords = eq_coords[0].transform_to('helioprojective')

        hp_tx = hp_coords[0].Tx.to_value(u.deg)
        eq_to_hp_tx = eq_to_hp_coords.Tx.to_value(u.deg)
        straddles_wrap = np.nonzero(np.abs(hp_tx - eq_to_hp_tx) > 350)
        for r, c in zip(*straddles_wrap):
            if hp_tx[r, c] > 170:
                hp_tx[r, c] -= 360
            else:
                eq_to_hp_tx[r, c] -= 360

        np.testing.assert_allclose(hp_tx, eq_to_hp_tx, atol=1e-12)
        np.testing.assert_allclose(hp_coords[0].Ty.to_value(u.deg), eq_to_hp_coords.Ty.to_value(u.deg), atol=1e-12)
        np.testing.assert_array_equal(hp_coords[1], eq_coords[1])


def test_load_trefoil_wcs():
    trefoil_wcs, trefoil_shape = load_trefoil_wcs()
    assert trefoil_shape == (4096, 4096)
    assert isinstance(trefoil_wcs, WCS)


@pytest.mark.parametrize("starting_rotation", [-140, -30, 0, 50, 100])
def test_back_and_forth_wcs_from_celestial(starting_rotation):
    date_obs = Time("2024-01-01T00:00:00", format='isot', scale='utc')
    sun_radec = get_sun(date_obs)
    wcs_celestial = WCS({"CRVAL1": sun_radec.ra.to(u.deg).value,
                         "CRVAL2": sun_radec.dec.to(u.deg).value,
                         "CRPIX1": 2047.5,
                         "CRPIX2": 2047.5,
                         "CDELT1": -0.0225,
                         "CDELT2": 0.0225,
                         "CUNIT1": "deg",
                         "CUNIT2": "deg",
                         "CTYPE1": "RA---AZP",
                         "CTYPE2": "DEC--AZP",
                         })
    pc = calculate_pc_matrix(starting_rotation, wcs_celestial.wcs.cdelt)
    wcs_celestial.wcs.pc = pc

    wcs_helio, p_angle = calculate_helio_wcs_from_celestial(wcs_celestial, date_obs, (10, 10))
    wcs_celestial_recovered = calculate_celestial_wcs_from_helio(wcs_helio, date_obs, (10, 10))

    assert np.allclose(wcs_celestial.wcs.crval, wcs_celestial_recovered.wcs.crval)
    assert np.allclose(wcs_celestial.wcs.crpix, wcs_celestial_recovered.wcs.crpix)
    assert np.allclose(wcs_celestial.wcs.pc, wcs_celestial_recovered.wcs.pc)
    assert np.allclose(wcs_celestial.wcs.cdelt, wcs_celestial_recovered.wcs.cdelt)
    assert list(wcs_celestial.wcs.ctype) == list(wcs_celestial_recovered.wcs.ctype)


@pytest.mark.parametrize("starting_rotation", [-140, -30, 0, 50, 100])
def test_back_and_forth_wcs_from_helio(starting_rotation):
    date_obs = Time("2024-01-01T00:00:00", format='isot', scale='utc')
    wcs_helio = WCS({"CRVAL1": 15.0,
                     "CRVAL2": 10.0,
                     "CRPIX1": 2047.5,
                     "CRPIX2": 2047.5,
                     "CDELT1": 0.0225,
                     "CDELT2": 0.0225,
                     "CUNIT1": "deg",
                     "CUNIT2": "deg",
                     "CTYPE1": "HPLN-ARC",
                     "CTYPE2": "HPLT-ARC"})
    pc = calculate_pc_matrix(starting_rotation, wcs_helio.wcs.cdelt)
    wcs_helio.wcs.pc = pc

    wcs_celestial = calculate_celestial_wcs_from_helio(wcs_helio.copy(), date_obs, (10, 10))
    wcs_helio_recovered, p_angle = calculate_helio_wcs_from_celestial(wcs_celestial.copy(), date_obs, (10, 10))

    assert np.allclose(wcs_helio.wcs.crval, wcs_helio_recovered.wcs.crval)
    assert np.allclose(wcs_helio.wcs.crpix, wcs_helio_recovered.wcs.crpix)
    assert np.allclose(wcs_helio.wcs.pc, wcs_helio_recovered.wcs.pc)
    assert np.allclose(wcs_helio.wcs.cdelt, wcs_helio_recovered.wcs.cdelt)
    assert list(wcs_helio_recovered.wcs.ctype) == list(wcs_helio.wcs.ctype)



import pytest
from prefect.testing.utilities import prefect_test_harness


@pytest.fixture(autouse=True, scope="session")
def prefect_test_fixture() -> None:
    """Test harness for prefect."""
    with prefect_test_harness(server_startup_timeout=60):
        yield



import pathlib

import numpy as np
import pytest
from pytest import fixture

from punchbowl.level1.destreak import correct_streaks, streak_correction_matrix

TEST_DIRECTORY = pathlib.Path(__file__).parent.resolve()


@pytest.mark.parametrize("diag, above, below", [(5, 4, 3), (5, 1, 1)])
def test_matrix_creation(diag, above, below):
    full_matrix = np.array(
        [
            [diag, above, above, above],
            [below, diag, above, above],
            [below, below, diag, above],
            [below, below, below, diag],
        ]
    )
    expected = np.linalg.inv(full_matrix)
    actual = streak_correction_matrix(4, diag, below, above)
    assert np.allclose(actual, expected)
    assert actual.shape == (4, 4)


@pytest.mark.parametrize(
    "size, exposure_time, readout_line_time, reset_line_time",
    [(5, 1, 2, 3), (100, 1, 2, 5)],
)
def test_blank_image_destreaking(
    size, exposure_time, readout_line_time, reset_line_time
):
    image = np.zeros((size, size))
    output = correct_streaks(image, exposure_time, readout_line_time, reset_line_time)
    assert output.shape == (size, size), "size should not change"
    assert np.allclose(output, 0), "an image of zeros destreaks to zeroes"


def test_not_square_image_destreaking():
    with pytest.raises(ValueError):
        image = np.zeros((50, 60))
        correct_streaks(image, 1, 2, 3)


def test_not_2dim_destreaking_errors():
    with pytest.raises(ValueError):
        image = np.zeros((50, 60, 70))
        correct_streaks(image, 1, 2, 3)


def test_not_numpy_image_destreaking_errors():
    with pytest.raises(TypeError):
        image = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        correct_streaks(image, 1, 2, 3)


def test_singular_matrix_errors():
    with pytest.raises(np.linalg.LinAlgError):
        image = np.zeros((3, 3))
        correct_streaks(image, 1, 1, 1)


# @pytest.mark.regression
# @fixture
# def used_correction_parameters() -> tuple:
#     exposure_time = 40
#     readout_line_time = 120 / 2148
#     reset_line_time = 120 / 2148
#     return exposure_time, readout_line_time, reset_line_time
#
#
# @pytest.mark.regression
# @fixture
# def expected_regression_test_input() -> np.ndarray:
#     """
#     This data is taken from the HotOp tests, specifically:
#         Nextcloud/23103_PUNCH_Data/WFI_PhaseC/PUNCH_WFI_EM_TBalance/archive_220116_HotOp_spot.fit
#     This is a calibrated 40 ms exposure following the steps in "Investigating destreaking WFI laboratory test data"
#     Jupyter notebook which can be found in the sdf at sdf/punchbowl/Level1/destreak_investigation_summary.ipynb
#
#     Returns
#     -------
#     np.ndarray
#         regression test input
#     """
#     path = TEST_DIRECTORY / "data/regression_input_destreak.npy"
#     return np.load(str(path))
#
#
# @pytest.mark.regression
# @fixture
# def expected_regression_test_output() -> np.ndarray:
#     path = TEST_DIRECTORY / "data/regression_output_destreak.npy"
#     return np.load(str(path))
#
#
# @pytest.mark.regression
# def test_regression(
#     expected_regression_test_input,
#     used_correction_parameters,
#     expected_regression_test_output,
# ):
#     test_output = correct_streaks(
#         expected_regression_test_input, *used_correction_parameters
#     )
#     assert np.allclose(test_output, expected_regression_test_output)



"""
===================
Decoding square root encoded data
===================

Data downlinked from the individual PUNCH spacecraft are usually square root encoded. This decoding is completed as part of the regular data processing pipeline for end-user data products. If using Level 0 manually, you may want to decode this data manually, as outlined here.
"""

# %%
# Data downlinked from the individual PUNCH spacecraft are usually square root encoded. This decoding is completed as part of the regular data processing pipeline for end-user data products. If using Level 0 manually, you may want to decode this data manually, as outlined here.

# %%
# Load libraries

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LogNorm

from punchbowl.data import load_ndcube_from_fits
from punchbowl.data.sample import PUNCH_DK4
from punchbowl.data.visualize import cmap_punch
from punchbowl.level1.sqrt import decode_sqrt

# %%
# We can begin by loading sample data into an NDCube object. Here we'll load a sample level 0 LED dark image from NFI. Note that if running with downloaded PUNCH data, replace the sample PUNCH_DK4 data below with a path pointing towards a FITS file.

# %%
cube = load_ndcube_from_fits(PUNCH_DK4)

# %%
# Note that for level 0 data the data is still square-root encoded, and will be unpacked in subsequent levels. You can manually decoded this data with the same pipeline function.

# %%
data_decoded = decode_sqrt(cube.data,
                      from_bits = 16,
                      to_bits = 11,
                      ccd_gain_right = 4.94,
                      ccd_gain_left = 4.89,
                      ccd_offset = 400,
                      ccd_read_noise = 17
                    )

# %%
# Now that we have this data square-root decoded, we can plot the image.

# %%
fig, ax = plt.subplots(figsize=(9.5, 7.5), subplot_kw={"projection":cube.wcs})

im = ax.imshow(data_decoded, cmap=cmap_punch, norm=LogNorm(vmax=450))

lon, lat = ax.coords
lat.set_major_formatter("dd")
lon.set_major_formatter("dd")
ax.set_facecolor("black")
ax.coords.grid(color="white", alpha=.25, ls="dotted")
ax.set_xlabel("Helioprojective longitude")
ax.set_ylabel("Helioprojective latitude")
ax.set_title("PUNCH Level 0 LED Image")
fig.colorbar(im, ax=ax, label="DN")
plt.show()

# %%



import os
import pathlib
import warnings
from collections.abc import Callable

from ndcube import NDCube

from punchbowl.data import load_ndcube_from_fits
from punchbowl.exceptions import (
    IncorrectPolarizationStateWarning,
    IncorrectTelescopeWarning,
    InvalidDataError,
    LargeTimeDeltaWarning,
    NoCalibrationDataWarning,
)
from punchbowl.prefect import punch_task


@punch_task
def correct_vignetting_task(data_object: NDCube, vignetting_path: str | pathlib.Path | Callable | None) -> NDCube:
    """
    Prefect task to correct the vignetting of an image.

    Vignetting is a reduction of an image's brightness or saturation toward the
    periphery compared to the image center, created by the optical path. The
    Vignetting Module will transform the data through a flat-field correction
    map, to cancel out the effects of optical vignetting created by distortions
    in the optical path. This module also corrects detector gain variation and
    offset.

    Correction maps will be 2048*2048 arrays, to match the input data, and
    built using the starfield brightness pattern. Mathematical Operation:

        I'_{i,j} = I_i,j * FF_{i,j}

    Where I_{i,j} is the number of counts in pixel i, j. I'_{i,j} refers to the
    modified value. FF_{i,j} is the small-scale flat field factor for pixel i,
    j. The correction mapping will take into account the orientation of the
    spacecraft and its position in the orbit.

    Uncertainty across the image plane is calculated using the modelled
    flat-field correction with stim lamp calibration data. Deviations from the
    known flat-field are used to calculate the uncertainty in a given pixel.
    The uncertainty is convolved with the input uncertainty layer to produce
    the output uncertainty layer.


    Parameters
    ----------
    data_object : PUNCHData
        data on which to operate

    vignetting_path : pathlib
        path to vignetting function to apply to input data

    Returns
    -------
    PUNCHData
        modified version of the input with the vignetting corrected

    """
    if vignetting_path is None:
        data_object.meta.history.add_now("LEVEL1-correct_vignetting", "Vignetting skipped")
        msg=f"Calibration file {vignetting_path} is unavailable, vignetting correction not applied"
        warnings.warn(msg, NoCalibrationDataWarning)
    else:
        if isinstance(vignetting_path, Callable):
            vignetting_function, vignetting_path = vignetting_path()
        else:
            if isinstance(vignetting_path, str):
                vignetting_path = pathlib.Path(vignetting_path)
            if not vignetting_path.exists():
                msg = f"File {vignetting_path} does not exist."
                raise InvalidDataError(msg)
            vignetting_function = load_ndcube_from_fits(vignetting_path, include_provenance=False)
        vignetting_function_date = vignetting_function.meta.astropy_time
        observation_date = data_object.meta.astropy_time
        if abs((vignetting_function_date - observation_date).to("day").value) > 14:
            msg = f"Calibration file {vignetting_path} contains data created greater than 2 weeks from the obsveration"
            warnings.warn(msg, LargeTimeDeltaWarning)
        if vignetting_function.meta["TELESCOP"].value != data_object.meta["TELESCOP"].value:
            msg = f"Incorrect TELESCOP value within {vignetting_path}"
            warnings.warn(msg, IncorrectTelescopeWarning)
        if vignetting_function.meta["OBSLAYR1"].value != data_object.meta["OBSLAYR1"].value:
            msg = f"Incorrect polarization state within {vignetting_path}"
            warnings.warn(msg, IncorrectPolarizationStateWarning)
        if vignetting_function.data.shape != data_object.data.shape:
            msg = f"Incorrect vignetting function shape within {vignetting_path}"
            raise InvalidDataError(msg)

        data_object.data[:, :] /= vignetting_function.data[:, :]
        data_object.uncertainty.array[:, :] /= vignetting_function.data[:, :]
        data_object.meta.history.add_now("LEVEL1-correct_vignetting",
                                         f"Vignetting corrected using {os.path.basename(str(vignetting_path))}")
    return data_object



import os
import pathlib

import numpy as np
import pytest
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube

from punchbowl.data import NormalizedMetadata, write_ndcube_to_fits
from punchbowl.level3.velocity import plot_flow_map, track_velocity

THIS_DIRECTORY = pathlib.Path(__file__).parent.resolve()


@pytest.fixture
def synthetic_data(tmpdir):
    """
    Create synthetic compressed FITS data from NDCube instances for testing.

    This fixture generates a list of file paths for FITS files containing random
    NDCube data. These files are written to a temporary directory and removed
    after the test session. Each file includes:
    - A 128x128 array of random data.
    - WCS metadata specifying helioprojective coordinates.
    - Normalized metadata according to the specified schema.
    - Uncertainty data initialized to zero.

    Returns:
        list of str: Paths to the generated FITS files.
    """
    files = []
    num_files = 5  # You can parameterize this if needed

    for i in range(num_files):
        # Generate random data
        data = np.random.rand(128, 128)

        # Define WCS for the NDCube
        wcs = WCS(naxis=2)
        wcs.wcs.ctype = ("HPLN-AZP", "HPLT-AZP")
        wcs.wcs.cunit = ("deg", "deg")
        wcs.wcs.cdelt = (0.02, 0.02)
        wcs.wcs.crpix = (64, 64)
        wcs.wcs.crval = (0, 24.75)
        wcs.array_shape = data.shape

        # Define metadata for the NDCube
        meta = NormalizedMetadata.load_template('PTM', '3')
        meta['DATE-OBS'] = "2024-01-01T00:00:00"
        meta['DATE-BEG'] = "2024-01-01T00:00:00"
        meta['DATE-END'] = "2024-01-01T00:00:00"
        meta['DATE-AVG'] = "2024-01-01T00:00:00"

        # Create NDCube
        uncertainty = StdDevUncertainty(np.zeros_like(data))
        cube = NDCube(data=data, wcs=wcs, meta=meta, uncertainty=uncertainty)

        # Write NDCube to a compressed FITS file
        file_path = os.path.join(str(tmpdir), f"file_{i}.fits")
        write_ndcube_to_fits(cube, str(file_path))
        files.append(str(file_path))

    return files


def test_shape_matching(synthetic_data):
    """Test that the output shape matches the expected configuration."""
    files = synthetic_data
    ycens = np.arange(7, 14.5, 0.5)
    result = track_velocity(files, ycens=ycens)

    assert isinstance(result, NDCube)
    assert result.data.shape[0] == len(ycens)


@pytest.mark.mpl_image_compare(style="default")
def test_wind_plot(synthetic_data):
    """Tests that wind plots are generated and  output to file."""
    files = synthetic_data
    ycens = np.arange(7, 14.5, 0.5)
    result = track_velocity(files, ycens=ycens)

    return plot_flow_map(None, result)


def test_no_nans_or_negatives(synthetic_data):
    """Test that the output does not contain NaNs or negative values."""
    files = synthetic_data
    result = track_velocity(files)

    assert not np.isnan(result.data).any(), "Data contains NaNs"
    assert (result.data >= 0).all(), "Data contains negative values"


def test_with_bad_data(tmpdir):
    """Test the function with intentionally bad data."""
    # Generate bad data (all NaNs)
    data = np.full((128, 128), np.nan)

    # Define WCS for the NDCube
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 64, 64
    wcs.wcs.crval = 0, 24.75

    # Define metadata for the NDCube
    meta = NormalizedMetadata.load_template('PTM', '3')
    meta['DATE-OBS'] = "2024-01-01T00:00:00"
    meta['DATE-BEG'] = "2024-01-01T00:00:00"
    meta['DATE-END'] = "2024-01-01T00:00:00"
    meta['DATE-AVG'] = "2024-01-01T00:00:00"

    # Create NDCube
    uncertainty = StdDevUncertainty(np.zeros_like(data))
    cube = NDCube(data=data, wcs=wcs, meta=meta, uncertainty=uncertainty)

    # Write NDCube to a compressed FITS file using your custom function
    file_path = os.path.join(str(tmpdir), "bad_file.fits")
    write_ndcube_to_fits(cube, file_path)

    with pytest.raises(ValueError):
        _ = track_velocity([str(file_path)])


def test_sample_radial_outflows(tmpdir):
    """Test the function with sample radial outflows."""
    files = []
    for i in range(5):
        radial_outflow_data = np.linspace(0, 1, 128)[:, None] * np.linspace(1, 0, 128)

        # Define WCS for the NDCube
        wcs = WCS(naxis=2)
        wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
        wcs.wcs.cunit = "deg", "deg"
        wcs.wcs.cdelt = 0.02, 0.02
        wcs.wcs.crpix = 64, 64
        wcs.wcs.crval = 0, 24.75

        # Define metadata for the NDCube
        meta = NormalizedMetadata.load_template('PTM', '3')
        meta['DATE-OBS'] = "2024-01-01T00:00:00"
        meta['DATE-BEG'] = "2024-01-01T00:00:00"
        meta['DATE-END'] = "2024-01-01T00:00:00"
        meta['DATE-AVG'] = "2024-01-01T00:00:00"

        # Create NDCube
        uncertainty = StdDevUncertainty(np.zeros_like(radial_outflow_data))
        cube = NDCube(data=radial_outflow_data, wcs=wcs, meta=meta, uncertainty=uncertainty)

        # Write NDCube to a compressed FITS file
        file_path = os.path.join(str(tmpdir), f"radial_outflow_file_{i}.fits")
        write_ndcube_to_fits(cube, file_path)
        files.append(str(file_path))

    result = track_velocity(files)

    assert isinstance(result, NDCube)
    assert result.data.mean() > 0  # Verify that there is a positive outflow signal



class PUNCHBowlError(Exception):
    """Base class for exceptions in punchbowl."""

class InvalidDataError(PUNCHBowlError):
    """Invalid data error."""

class InvalidHeaderError(PUNCHBowlError):
    """Header is not properly formatted."""

class MissingMetadataError(PUNCHBowlError):
    """Metadata missing for processing."""

class IncorrectFileCountError(PUNCHBowlError):
    """Wrong number of files passed in."""


class PUNCHBowlWarning(Warning):
    """Base class for warnings in punchbowl."""

class LargeTimeDeltaWarning(PUNCHBowlWarning):
    """Large time delta warning between datasets."""

class NoCalibrationDataWarning(PUNCHBowlWarning):
    """Calibration skipped no valid calibration data available."""

class ExtraMetadataWarning(PUNCHBowlWarning):
    """Extra metadata found but ignored."""

class IncorrectPolarizationStateWarning(PUNCHBowlWarning):
    """Mismatched polarization state detected but ignored."""

class IncorrectTelescopeWarning(PUNCHBowlWarning):
    """Mismatched telescope detected but ignored."""

class DataValueWarning(PUNCHBowlWarning):
    """Data values may exceed expected values."""



import numpy as np
from ndcube import NDCube
from prefect import get_run_logger
from scipy.ndimage import gaussian_filter
from skimage.morphology import binary_dilation

from punchbowl.data import load_ndcube_from_fits
from punchbowl.prefect import punch_task


def run_zspike(
    data: np.ndarray,
    uncertainty: np.ndarray,
    threshold: float = 4,
    required_yes: int = 6,
    veto_limit: int = 2,
    diff_method: str = "sigma",
    dilation: int = 0,
    blur_scale: float = 1.0,
    blur_threshold: float | None = 0.15,
    index_of_interest: int | None = None,
    ) -> np.ndarray:
    """
    Identify bright structures in temporal series of images.

    Given a time sequence of images, identify "spikes" that exceed a
    threshold in a single frame.

    Diffuse bright structures over the background F-corona are
    identified and marked in the data using in-band "bad value" marking
    (which is supported by the FITS standard). Data marking follows the
    existing ZSPIKE temporal despiking algorithm to identify auroral
    transients.

    This is a voting algorithm based on ZSPIKE.PRO from the 1990s,
    which is available in the solarsoft distribution from Lockheed Martin.

    ZSPIKE was originally used to identify cosmic rays, and
    was adopted on STEREO for on-board despiking during exposure accumulation.
    For this application ZSPIKE is ideal because it does not rely on the
    spatial structure of spikes, only their temporal structure. Both cosmic
    rays and, if present, high-altitude aurora are transient and easily
    detected with ZSPIKE.

    The algorithm assembles "votes" from the images
    surrounding each one in a stream, to determine whether a particular pixel
    is a good candidate for a temporal spike. If the pixel is sufficiently
    bright compared to its neighbors in time, it is marked bad. "Bad values"
    are stored in the DRP for file quality marking outlined in Module Quality
    Marking.

    There are two methods of identifying if a pixel is above a given threshold,
    and therefore considered a spike. diff_method='abs' or 'sigma'.

    diff_method 'abs' represents the absolute difference, and is the default.
    If set, this is an absolute difference, in DN, required for a pixel to
    'vote' its central value.  If the central value is this much higher than a
    given voting value, then the central value is voted to be a spike.  If
    it's this much lower, the veto count is incremented.

    If diff_method 'sigma' is set if, then each pixel is treated as a
    time series and the calculated sigma (RMS variation from the mean) of
    the timeseries is used to calculate a difference threshold at each
    location.

    The threshold is the value over which a pixel is voted as a spike. The threshold
    should be different depending diff_method.

    Parameters
    ----------
    data : np.ndarray
        data to operate on - this is expected to have an odd number of frames, or
        a frame_of_interest should be inserted.

    uncertainty: np.ndarray
        data to operate on - this is expected to have an odd number of frames, or
        a frame_of_interest should be inserted.

    threshold : float
        This is the threshold over which a pixel is voted as a spike.

    required_yes: int
        (default is 4) - number of 'voting' frames that must vote the
        value is a spike, for it to be marked as such.

    veto_limit: int
        (default is 2) - number of 'voting' frames that must vote NO to veto
        marking the central value as a spike.

    diff_method: str
        This is the method by which the threshold is set. 'abs' treats each pixel
        independently and finds the absolute difference, and 'sigma' treats each
        corresponding pixel as a time series and the calculated RMS variation
        from the mean of the timeseries is used to calculate a difference at each
        location.

    dilation: int
        If nonzero, this is the number of times to morphologically dilate pixels marked as bright structures.

    blur_scale: float
        The scale of the gaussian kernel used for convolving with the mask for blurring.

    blur_threshold: float
        The threshold value used to cast a blurred mask array back to a boolean data type.

    index_of_interest : int
        if you have an even number of frames, or you don't want to find spikes
        in the center frame, the frame_of_interest will be used. index_of_interest
        starts from 0, therefore the center frame from 21 frames will be 10.

    Returns
    -------
    np.ndarray
        a frame matching the dimensions of the frame of interest with True flagging
        bad pixels, and False flagging good data points

    """
    if len(data.shape) != 3:
        msg = "`data` must be a 3-D array"
        raise ValueError(msg)

    # test if odd number of frames, or if a frame of interest has been included
    z_shape = np.shape(data[:, 0, 0])

    if z_shape[0] % 2 == 0:
        if index_of_interest is None:
            msg = "Number of frames in `data` must be odd or have `frame_of_interest` set."
            raise ValueError(msg)
    elif index_of_interest is None:
        index_of_interest = z_shape[0] // 2

    frame_of_interest = data[index_of_interest, :, :]
    frame_of_interest_uncertainty = uncertainty[index_of_interest, :, :]

    voters_array = np.delete(data, index_of_interest, 0)
    voters_array_uncertainty = np.delete(uncertainty, index_of_interest, 0)

    # create a reference cube the same dimensions as the voter cube, filled with frame of interest
    reference_cube = np.stack([frame_of_interest for _ in range(voters_array.shape[0])], axis=0)
    difference_array = np.abs(reference_cube - voters_array)

    if diff_method == "abs":
        threshold_array = np.full(voters_array.shape, threshold)
    elif diff_method == "sigma":
        threshold_array = threshold * np.nanstd(voters_array, axis=0, where=voters_array_uncertainty < 1.0)
    else:
        msg = f"A `diff_method` of `sigma` or `abs` is expected. Found diff_method={diff_method}."
        raise ValueError(msg)

    yes_vote_count = np.sum(difference_array > threshold_array, axis=0)
    no_vote_count = np.full(frame_of_interest.shape, voters_array.shape[0]) - yes_vote_count
    flagged_features_array = yes_vote_count > required_yes

    # if the number of no votes exceeds a veto limit, then veto
    veto_mask = no_vote_count > veto_limit

    flagged_features_array[veto_mask] = False

    # if input data already has an unbounded uncertainty, create a True flag
    flagged_features_array = np.where(np.isinf(frame_of_interest_uncertainty), True, flagged_features_array)

    # expand flags by dilating to remove holes or fuzzy edges
    for _ in range(dilation):
        binary_dilation(flagged_features_array, out=flagged_features_array)

    # blur and threshold the mask
    if blur_scale is not None:
        flagged_features_array = gaussian_filter(flagged_features_array.astype(float),
                                                 sigma=blur_scale) > blur_threshold

    return flagged_features_array


@punch_task
def identify_bright_structures_task(
    data: NDCube | None,
    voter_filenames: list[str],
    threshold: float = 4,
    required_yes: int = 6,
    veto_limit: int = 2,
    diff_method: str = "sigma",
    dilation: int = 0,
) -> NDCube | None:
    """Prefect task to perform bright structure identification."""
    logger = get_run_logger()
    logger.info("identify_bright_structures_task started")

    if data is None:
        logger.info("Identify bright structures skipped since data is None")
        return data

    if len(voter_filenames) == 0:
        logger.info("Identify bright structures skipped since no voters provided")
        data.meta.history.add_now("LEVEL2-bright_structure",
                                  "Identify bright structures skipped since no voters provided")
        return data

    # construct voter cube
    voters, voters_uncertainty = [], []
    for voter_filename in voter_filenames:
        this_punchdata = load_ndcube_from_fits(voter_filename)
        voters.append(this_punchdata.data)
        voters_uncertainty.append(this_punchdata.uncertainty)

    # add frame of interest
    voters.append(data.data)
    voters_uncertainty.append(data.uncertainty)

    # apply find spikes
    spike_mask = run_zspike(
        data=np.array(voters),
        uncertainty=np.array(voters_uncertainty),
        threshold=threshold,
        required_yes=required_yes,
        veto_limit=veto_limit,
        diff_method=diff_method,
        dilation=dilation,
        index_of_interest=-1)

    # add the uncertainty to the output punch data object
    data.uncertainty.array[spike_mask] = np.inf

    logger.info("identify_bright_structures_task ended")
    data.meta.history.add_now("LEVEL2-bright_structures",
                              "bright structure identification completed")

    return data



import os
from datetime import UTC, datetime

import numpy as np
from ndcube import NDCube

from punchbowl.data import load_ndcube_from_fits, write_ndcube_to_fits
from punchbowl.prefect import punch_task


def validate_image_is_square(image: np.ndarray) -> None:
    """Check that the input array is square."""
    if not isinstance(image, np.ndarray):
        msg = f"Image must be of type np.ndarray. Found: {type(image)}."
        raise TypeError(msg)
    if len(image.shape) != 2:
        msg = f"Image must be a 2-D array. Input has {len(image.shape)} dimensions."
        raise ValueError(msg)
    if not np.equal(*image.shape):
        msg = f"Image must be square. Found: {image.shape}."
        raise ValueError(msg)


@punch_task
def output_image_task(data: NDCube, output_filename: str) -> None:
    """
    Prefect task to write an image to disk.

    Parameters
    ----------
    data : NDCube
        data that is to be written
    output_filename : str
        where to write the file out

    Returns
    -------
    None

    """
    output_dir = os.path.dirname(output_filename)
    if output_dir and not os.path.isdir(output_dir):
        os.makedirs(output_dir)
    write_ndcube_to_fits(data, output_filename)


@punch_task
def load_image_task(input_filename: str, include_provenance: bool = True, include_uncertainty: bool = True) -> NDCube:
    """
    Prefect task to load data for processing.

    Parameters
    ----------
    input_filename : str
        path to file to load
    include_provenance : bool
        whether to load the provenance layer
    include_uncertainty : bool
        whether to load the uncertainty layer

    Returns
    -------
    NDCube
        loaded version of the image

    """
    return load_ndcube_from_fits(
        input_filename, include_provenance=include_provenance, include_uncertainty=include_uncertainty)


def average_datetime(datetimes: list[datetime]) -> datetime:
    """Compute average datetime from a list of datetimes."""
    timestamps = [dt.timestamp() for dt in datetimes]
    average_timestamp = sum(timestamps) / len(timestamps)
    return datetime.fromtimestamp(average_timestamp).astimezone(UTC)


def _zvalue_from_index(arr, ind):  # noqa: ANN202, ANN001
    """
    Do math.

    Private helper function to work around the limitation of np.choose() by employing np.take().
    arr has to be a 3D array
    ind has to be a 2D array containing values for z-indicies to take from arr
    See: http://stackoverflow.com/a/32091712/4169585
    This is faster and more memory efficient than using the ogrid based solution with fancy indexing.
    """
    # get number of columns and rows
    _, n_rows, n_cols = arr.shape

    # get linear indices and extract elements with np.take()
    idx = n_cols*n_rows*ind + n_cols*np.arange(n_rows)[:,None] + np.arange(n_cols)
    return np.take(arr, idx)


def nan_percentile(arr: np.ndarray, q: list[float] | float) -> np.ndarray:
    """Calculate the nan percentile faster of a 3D cube."""
    # np.nanpercentile is slow so use this: https://krstn.eu/np.nanpercentile()-there-has-to-be-a-faster-way/

    # valid (non NaN) observations along the first axis
    is_good = np.isfinite(arr)
    n_valid_obs = np.sum(is_good, axis=0)
    # replace NaN with maximum
    arr = arr.copy()
    arr[~is_good] = np.nanmax(arr)
    # sort - former NaNs will move to the end
    arr = np.sort(arr, axis=0)

    # loop over requested quantiles
    qs = [q] if isinstance(q, float | int) else q

    result = np.empty((len(qs), *arr.shape[1:]))
    for i, quant in enumerate(qs):
        # desired position as well as floor and ceiling of it
        k_arr = (n_valid_obs - 1) * (quant / 100)
        f_arr = np.floor(k_arr).astype(np.int32)
        c_arr = np.ceil(k_arr).astype(np.int32)
        fc_equal_k_mask = f_arr == c_arr

        # linear interpolation (like numpy percentile) takes the fractional part of desired position
        floor_val = _zvalue_from_index(arr=arr, ind=f_arr) * (c_arr - k_arr)
        ceil_val = _zvalue_from_index(arr=arr, ind=c_arr) * (k_arr - f_arr)

        quant_arr = floor_val + ceil_val
        # if floor == ceiling take floor value
        quant_arr[fc_equal_k_mask] = _zvalue_from_index(arr=arr, ind=k_arr.astype(np.int32))[fc_equal_k_mask]

        result[i] = quant_arr

    result[:, n_valid_obs == 0] = np.nan

    return result



import os
from collections.abc import Callable

import numexpr as ne
import numpy as np
from ndcube import NDCube

from punchbowl.data import load_ndcube_from_fits
from punchbowl.prefect import punch_task


def create_coefficient_image(flat_coefficients: np.ndarray, image_shape: tuple) -> np.ndarray:
    """
    Given a set of coefficients that should apply for every pixel, convert them to the required format.

    Parameters
    ----------
    flat_coefficients : np.ndarray
        A one-dimensional list of coefficients that should apply to every pixel in the image.
        Coefficients should be ordered from the highest power to lowest as expected in `photometric_calibration`, e.g.
        f(i,j) = a+b*IMG[i,j]+c*IMG[i,j]^2 would have flat_coefficients of [c, b, a]
    image_shape : tuple
        A tuple of the shape of the image that will be calibrated using `photometric_calibration`

    Returns
    -------
    np.ndarray
        An image of coefficients that apply to every pixel as expected by `photometric_calibration`

    """
    return np.stack([np.ones(image_shape) * coeff for coeff in flat_coefficients], axis=0)


def create_constant_quartic_coefficients(img_shape: tuple) -> np.ndarray:
    """
    Create a constant coefficients image that preserves the original values, b = 1 and all other coefficients are 0.

    Parameters
    ----------
    img_shape : tuple[Int]
        size of the image to create the coefficients for

    Returns
    -------
    np.ndarray
        An image of coefficients that apply to every pixel as expected by `photometric_calibration`

    """
    return create_coefficient_image(np.array([0, 0, 0, 1, 0]), img_shape)


def photometric_calibration(image: np.ndarray, coefficient_image: np.ndarray) -> np.ndarray:
    """
    Compute a non-linear photometric calibration of PUNCH images.

    Parameters
    ----------
    image : np.ndarray
        Image to be corrected.

    coefficient_image : np.ndarray
        Frame containing uncertainty values.
        The first two dimensions are the spatial dimensions of the image.
        The last dimension iterates over the powers of the coefficients, starting with index 0 being the highest power
        and counting down.

    Returns
    -------
    np.ndarray
        a photometrically corrected frame

    Notes
    -----
    Each instrument is subject to an independent non-linear photometric response,
    which needs to be corrected. The module converts from raw camera digitizer
    number (DN) to photometric units at each pixel. Each pixel is replaced with
    the corresponding value of the quartic polynomial in the current calibration file data
    product for that particular camera.

    A quartic polynomial is applied as follows:

    .. math:: X_{i,j} = a_{i,j}+b_{i,j}*DN_{i,j}+c_{i,j}*DN_{i,j}^2+d_{i,j}*DN_{i,j}^3+e_{i,j}*DN_{i,j}^4

    for each pixel in the detector. Each quantity (a, b, c, d, e) is a function
    of pixel location (i,j), and is generated using dark current and Stim lamp
    maps. a = offset (dark and the bias). b, c, d, e = higher order terms.
    Specifically ``coefficient_image[:,i,j] = [e, d, c, b, a]`` (highest order terms first)

    As each pixel is independent, a quartic fit calibration file of
    dimensions 2k*2k*5 is constructed, with each layer containing one of the five
    polynomial coefficients for each pixel.

    Examples
    --------
    >>> punch_image = np.ones((100,100))
    >>> coefficient_image = create_coefficient_image(np.array([0, 0, 0, 1, 0]), punch_image.shape)
    >>> data = photometric_calibration(punch_image, coefficient_image)

    """
    # inspect dimensions
    if len(image.shape) != 2:
        msg = "`image` must be a 2-D image"
        raise ValueError(msg)

    if len(coefficient_image.shape) != 3:
        msg = "`coefficient_image` must be a 3-D image"
        raise ValueError(msg)

    if coefficient_image.shape[1:] != image.shape:
        msg = "`coefficient_image` and `image` must have the same shape`"
        raise ValueError(msg)

    # find the number of quartic fit coefficients
    num_coefficients = coefficient_image.shape[0]
    # Here we compute sum(coefficient_image[i] * image ** (num_coefficients - i - 1))
    # This is a good domain for numexpr, and using it speeds up the full L1 flow by ~10%. But because numexpr doesn't
    # support indexing, we have to pre-define variables for each slice of coefficient_image. And because we're handling
    # a dynamic number of coefficients, we have to build this up programmatically.
    # This will contain the pieces of the expression (each term in the summation)
    pieces = []
    # This will contain all the variables that go into the numexpr expression
    inputs = {"image": image}
    for i in range(num_coefficients):
        # Define a "variable" for this slice
        inputs[f"a{i}"] = coefficient_image[i]
        # Write this part of the summation
        pieces.append(f"a{i} * image ** {num_coefficients - i - 1}")
    return ne.evaluate(" + ".join(pieces), local_dict=inputs)


@punch_task
def perform_quartic_fit_task(data_object: NDCube, quartic_coefficients_path: str | Callable | None = None) -> NDCube:
    """
    Prefect task to perform the quartic fit calibration on the data.

    Parameters
    ----------
    data_object : NDCube
        a data object that needs calibration
    quartic_coefficients_path: Optional[str]
        path to a  cube of coefficients as produced by `create_coefficients_image` or `create_ones_coefficients_image`,
        skips correction if it is None

    Returns
    -------
    NDCube
        modified version of the input with the quartic fit correction applied

    See Also
    --------
    photometric_calibration

    """
    if quartic_coefficients_path is not None:
        if isinstance(quartic_coefficients_path, Callable):
            quartic_coefficients, quartic_coefficients_path = quartic_coefficients_path()
        else:
            quartic_coefficients = load_ndcube_from_fits(quartic_coefficients_path)
        new_data = photometric_calibration(data_object.data, quartic_coefficients.data)
        data_object.data[...] = new_data[...]

        data_object.meta.history.add_now(
            "LEVEL1-quartic_fit",
            f"Quartic fit correction completed with {os.path.basename(quartic_coefficients_path)}",
        )
    else:
        data_object.meta.history.add_now("LEVEL1-quartic_fit", "Quartic fit correction skipped since path is empty")

    return data_object



import os
import pathlib
import warnings

import numpy as np
from astropy.time import Time
from ndcube import NDCube

from punchbowl.data import load_ndcube_from_fits
from punchbowl.exceptions import (
    IncorrectPolarizationStateWarning,
    IncorrectTelescopeWarning,
    InvalidDataError,
    LargeTimeDeltaWarning,
)
from punchbowl.prefect import punch_task
from punchbowl.util import nan_percentile


def estimate_stray_light(filepaths: [str], percentile: float = 3) -> np.ndarray:
    """Estimate the fixed stray light pattern using a percentile."""
    data = np.array([load_ndcube_from_fits(path).data for path in filepaths])
    return nan_percentile(data, percentile)


@punch_task
def remove_stray_light_task(data_object: NDCube, stray_light_path: pathlib) -> NDCube:
    """
    Prefect task to remove stray light from an image.

    Stray light is light in an optical system which was not intended in the
    design.

    The PUNCH instrument stray light will be mapped periodically as part of the
    ongoing in-flight calibration effort. The stray light maps will be
    generated directly from the L0 and L1 science data. Separating instrumental
    stray light from the F-corona. This has been demonstrated with SOHO/LASCO
    and with STEREO/COR2 observations. It requires an instrumental roll to hold
    the stray light pattern fixed while the F-corona rotates in the field of
    view. PUNCH orbital rolls will be used to create similar effects.

    Uncertainty across the image plane is calculated using a known stray light
    model and the difference between the calculated stray light and the ground
    truth. The uncertainty is convolved with the input uncertainty layer to
    produce the output uncertainty layer.


    Parameters
    ----------
    data_object : NDCube
        data to operate on

    stray_light_path: pathlib
        path to stray light model to apply to data

    Returns
    -------
    NDCube
        modified version of the input with the stray light removed

    """
    if stray_light_path is None:
        data_object.meta.history.add_now("LEVEL1-remove_stray_light", "Stray light correction skipped")
    elif not stray_light_path.exists():
        msg = f"File {stray_light_path} does not exist."
        raise InvalidDataError(msg)
    else:
        stray_light_model = load_ndcube_from_fits(stray_light_path)

        stray_light_model_date = Time(stray_light_model.meta["DATE-OBS"].value)
        observation_date = Time(data_object.meta["DATE-OBS"].value)
        if abs((stray_light_model_date - observation_date).to("day").value) > 14:
            msg=f"Calibration file {stray_light_path} contains data created greater than 2 weeks from the obsveration"
            warnings.warn(msg,LargeTimeDeltaWarning)

        if stray_light_model.meta["TELESCOP"].value != data_object.meta["TELESCOP"].value:
            msg=f"Incorrect TELESCOP value within {stray_light_path}"
            warnings.warn(msg, IncorrectTelescopeWarning)
        elif stray_light_model.meta["OBSLAYR1"].value != data_object.meta["OBSLAYR1"].value:
            msg=f"Incorrect polarization state within {stray_light_path}"
            warnings.warn(msg, IncorrectPolarizationStateWarning)
        elif stray_light_model.data.shape != data_object.data.shape:
            msg = f"Incorrect stray light function shape within {stray_light_path}"
            raise InvalidDataError(msg)
        else:
            data_object.data[:, :] -= stray_light_model.data[:, :]
            data_object.uncertainty.array[...] -= stray_light_model.data[:, :]
            data_object.meta.history.add_now("LEVEL1-remove_stray_light",
                                             f"stray light removed with {os.path.basename(str(stray_light_model))}")
    return data_object



"""
Module for downloading sample data files.

This module provides the following sample data files.  When a sample shortname
is accessed, the corresponding file is downloaded if needed.  All files can be
downloaded by calling :func:`~punchbowl.data.sample.download_all`.

Code adapted from SunPy.
"""
from pathlib import Path
from urllib.parse import urljoin

import parfive
from sunpy import log
from sunpy.util.config import _is_writable_dir
from sunpy.util.parfive_helpers import Downloader

_BASE_URLS = (
    "https://data.boulder.swri.edu/lowder/PUNCH/sample/"
)

_SAMPLE_DATA = {
    "PUNCH_DK4": "PUNCH_L0_DK4_20250414154536_v1.fits",
    "PUNCH_PAM": "PUNCH_L3_PAM_20240620000000.fits",
    "PUNCH_PAN": "PUNCH_L3_PAN_20240620000000.fits",
    "PUNCH_PNN": "PUNCH_L3_PNN_20240620000000.fits",
    "PUNCH_PTM": "PUNCH_L3_PTM_20240620000000.fits",
    "QUICKPUNCH_WQM": "PUNCH_LQ_WQM_20230704012000.fits",
    "QUICKPUNCH_NQN": "PUNCH_LQ_NQN_20230704022400.fits",
}

_SAMPLE_FILES = {v: k for k, v in _SAMPLE_DATA.items()}


def get_and_create_sample_dir() -> Path:
    """Get the config of download directory and create one if not present."""
    sample_dir = Path("PUNCH_SAMPLE").expanduser().resolve()
    if not _is_writable_dir(sample_dir):
        raise RuntimeError(f'Could not write to SunPy sample data directory="{sample_dir}"')

    return sample_dir


def _download_sample_data(base_url:str, sample_files:list, overwrite:bool) -> parfive.Results:
    """
    Download a list of files.

    Parameters
    ----------
    base_url : str
        Base URL for each file.
    sample_files : list of tuples
        List of tuples that are (URL_NAME, SAVE_NAME).
    overwrite : bool
        Will overwrite a file on disk if True.

    Returns
    -------
    `parfive.Results`
        Download results. Will behave like a list of files.

    """
    dl = Downloader(overwrite=overwrite, progress=True)

    for url_file_name, fname in sample_files:
        url = urljoin(base_url, url_file_name)
        dl.enqueue_file(url, filename=fname)

    return dl.download()


def _retry_sample_data(results: parfive.Results, new_url_base:str) -> parfive.Results:
    """
    Retry failed downloads.

    Parameters
    ----------
    results : `parfive.Results`
        Download results.
    new_url_base : str
        Base URL for each file.

    Returns
    -------
    `parfive`
        Download results. Will behave like a list of files.

    """
    dl = Downloader(overwrite=True, progress=True)

    for err in results.errors:
        file_name = err.url.split("/")[-1]
        log.debug(
            f"Failed to download {_SAMPLE_FILES[file_name]} from {err.url}: {err.exception}")
        # Update the url to a mirror and requeue the file.
        new_url = urljoin(new_url_base, file_name)
        log.debug(f"Attempting redownload of {_SAMPLE_FILES[file_name]} using {new_url}")
        dl.enqueue_file(new_url, filename=err.filepath_partial)

    extra_results = dl.download()

    # Make a new results object which contains all the successful downloads
    # from the previous results object and this retry, and all the errors from
    # this retry.
    new_results = results + extra_results
    new_results._errors = extra_results._errors  # noqa: SLF001
    return new_results


def _handle_final_errors(results:parfive.Results) -> None:
    for err in results.errors:
        file_name = err.url.split("/")[-1]
        log.debug(f"Failed to download {_SAMPLE_FILES[file_name]} from {err.url}: {err.exception}",
                  )
        log.error(
            f"Failed to download {_SAMPLE_FILES[file_name]} from all mirrors,"
            "the file will not be available.",
        )


def _get_sampledata_dir() -> Path:
    return Path(get_and_create_sample_dir())


def _get_sample_files(filename_list: list, no_download:bool=False, force_download:bool=False) -> list[Path]:
    """
    Return a list of disk locations for a list of sample data filenames, downloading sample data files as needed.

    Parameters
    ----------
    filename_list : `list` of `str`
        List of filenames for sample data
    no_download : `bool`
        If ``True``, do not download any files, even if they are not present.
        Default is ``False``.
    force_download : `bool`
        If ``True``, download all files, and overwrite any existing ones.
        Default is ``False``.

    Returns
    -------
    `list` of `pathlib.Path`
        List of disk locations corresponding to the list of filenames.  An entry
        will be ``None`` if ``no_download == True`` and the file is not present.

    Raises
    ------
    RuntimeError
        Raised if any of the files cannot be downloaded from any of the mirrors.

    """
    sampledata_dir = _get_sampledata_dir()

    fullpaths = [sampledata_dir/fn for fn in filename_list]

    if no_download:
        fullpaths = [fp if fp.exists() else None for fp in fullpaths]
    else:
        to_download = zip(filename_list, fullpaths, strict=False)
        if not force_download:
            to_download = [(fn, fp) for fn, fp in to_download if not fp.exists()]

        if to_download:
            results = _download_sample_data(_BASE_URLS, to_download, overwrite=force_download)

            # Try the other mirrors for any download errors
            if results.errors:
                _handle_final_errors(results)
                raise RuntimeError

    return fullpaths


# Add a table row to the module docstring for each sample file
for _keyname, _filename in sorted(_SAMPLE_DATA.items()):
    __doc__ += f"   * - ``{_keyname}``\n     - {_filename}\n"


# file_dict and file_list are not normal variables; see __getattr__() below
__all__ = sorted(_SAMPLE_DATA.keys()) + ["download_all", "file_dict", "file_list"]  # noqa: F822, PLE0605, RUF005


# See PEP 562 (https://peps.python.org/pep-0562/) for module-level __dir__()
def __dir__():  # noqa: ANN202
    return __all__


# See PEP 562 (https://peps.python.org/pep-0562/) for module-level __getattr__()
def __getattr__(name):  # noqa: ANN001, ANN202
    if name in _SAMPLE_DATA:
        return _get_sample_files([_SAMPLE_DATA[name]])[0]
    if name == "file_dict":
        return dict(sorted(zip(_SAMPLE_DATA.keys(),
                               _get_sample_files(_SAMPLE_DATA.values(), no_download=True), strict=False)))
    if name == "file_list":
        return [v for v in __getattr__("file_dict").values() if v]
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


def download_all(force_download:bool=False) -> list[Path]:
    """
    Download all sample data at once that has not already been downloaded.

    Parameters
    ----------
    force_download : `bool`
        If ``True``, files are downloaded even if they already exist.  Default is
        ``False``.

    Returns
    -------
    `list` of `pathlib.Path`
        List of filepaths for sample data

    """
    return _get_sample_files(_SAMPLE_DATA.values(), force_download=force_download)



import numpy as np
from ndcube import NDCube
from threadpoolctl import threadpool_limits

from punchbowl.prefect import punch_task
from punchbowl.util import validate_image_is_square


def streak_correction_matrix(
    n: int, exposure_time: float, readout_line_time: float, reset_line_time: float,
) -> np.ndarray:
    """
    Compute a matrix used in correcting streaks in PUNCH images.

    Computes the inverse of a matrix of size n where the major diagonal
    contains the value exposure_time, the lower triangle contains readout_line_time
    and the upper triangle contains the reset_line_time.
    i.e. X[i,i]=diag, X[0:i-1,i]=below, X[0,i+1:n-1]=above

    Adapted from solarsoft sc_inverse

    Parameters
    ----------
    n : int
        size of the matrix (n x n)
    exposure_time : float
        the exposure time, i.e. value on the diagonal of the matrix
    readout_line_time : float
        the readout line time, i.e. value in the lower triangle
    reset_line_time : float
        the reset line time, i.e. value in the upper triangle

    Returns
    -------
    np.ndarray
        value of specified streak correction array

    Raises
    ------
    np.linalg.LinAlgError: Singular matrix
        Matrix isn't invertible

    Notes
    -----
    As long as the units are consistent, this should work. For PUNCH, we use milliseconds.

    Examples
    --------
    >>> streak_correction_matrix(3, 1, 2, 3)
    array([[-0.38461538,  0.23076923,  0.46153846],
           [ 0.30769231, -0.38461538,  0.23076923],
           [ 0.15384615,  0.30769231, -0.38461538]])

    """
    lower = np.tril(np.ones((n, n)) * readout_line_time, -1)
    upper = np.triu(np.ones((n, n)) * reset_line_time, 1)
    diagonal = np.diagflat(np.ones(n) * exposure_time)
    full_matrix = lower + upper + diagonal
    return np.linalg.inv(full_matrix)


def correct_streaks(
    image: np.ndarray,
    exposure_time: float,
    readout_line_time: float,
    reset_line_time: float,
    max_workers: int | None = None,
) -> np.ndarray:
    """
    Corrects an image for streaks.

    Parameters
    ----------
    image : np.ndarray (2D)
        image to be corrected
    exposure_time : float
        exposure time in consistent units (e.g. milliseconds) with readout_line_time and reset_line time
    readout_line_time : float
        time to read out a line in consistent units (e.g. milliseconds) with exposure_time and reset_line time
    reset_line_time : float
        time to reset CCD in consistent units (e.g. milliseconds) with readout_line_time and exposure_time
    max_workers : int
        the number of worker threads to use. If None, defaults to the system CPU count

    Returns
    -------
    np.ndarray
        a streak-corrected image

    Raises
    ------
    ValueError
        If the image is not 2D or not square
    TypeError
        If the image is not a numpy array
    np.linalg.LinAlgError: Singular matrix
        Matrix isn't invertible

    Examples
    --------
    >>> correct_streaks(np.arange(9).reshape(3,3), 1, 2, 3)
    array([[ 3.46153846,  3.76923077,  4.07692308],
       [ 0.23076923,  0.38461538,  0.53846154],
       [-1.38461538, -1.30769231, -1.23076923]])

    """
    validate_image_is_square(image)
    with threadpool_limits(max_workers):
        correction_matrix = streak_correction_matrix(image.shape[0], exposure_time, readout_line_time, reset_line_time)
        return correction_matrix @ image


@punch_task
def destreak_task(data_object: NDCube,
                  exposure_time: float = 1.0,
                  readout_line_time: float = 0.1,
                  reset_line_time: float = 0.1,
                  max_workers: int | None = None) -> NDCube:
    """Prefect task to destreak an image."""
    new_data = correct_streaks(
        data_object.data, exposure_time, readout_line_time, reset_line_time, max_workers=max_workers)
    data_object.data[...] = new_data[...] * exposure_time

    # data_object.uncertainty.array[...] = correct_streaks(data_object.uncertainty.array,
    #                                                 exposure_time, readout_line_time, reset_line_time) * exposure_time

    data_object.meta.history.add_now("LEVEL1-destreak", "image destreaked")
    data_object.meta.history.add_now("LEVEL1-destreak", f"exposure_time={exposure_time}")
    data_object.meta.history.add_now("LEVEL1-destreak", f"readout_line_time={readout_line_time}")
    data_object.meta.history.add_now("LEVEL1-destreak", f"reset_line_time={reset_line_time}")
    return data_object
